'''
Created on Mar 29, 2011

@author: ari
'''
from vector import *
import time

#helper functions
def list_intersect(a, b):
    return [x for x in a if x in b]

class Vertex:
    max_out_of_plane = 0.1
    
    def __init__(self, pos):
        self.pos = pos
        self.faces=[]
        self.edges=[]
        self.flag = False
        
    def set_edges(self, edges):
        self.edges = edges
        
    def set_faces(self, faces):
        self.faces = faces

    def add_face(self, f):
        self.faces.append(f)

    def add_edge(self, e):
        self.edges.append(e)

    def remove_edge(self, e):
        if e in self.edges:
            self.edges.remove(e)

    def remove_face(self, f):
        if f in self.faces:
            self.faces.remove(f)

    def clear(self):
        for f in self.faces:
            f.remove_vertex(self)
        for e in self.edges:
            e.remove_vertex(self)

    def neighbors(self):   
        v_neighbors =[]     
        for e in self.edges:
            if e.vertices[0]==self:
                v_neighbors.append(e.vertices[1])
            else:
                v_neighbors.append(e.vertices[0])
        return v_neighbors

    def normal(self):
        adjacent_normal_sum = Vector([0,0,0])
        for neigh in self.faces:
            adjacent_normal_sum += neigh.normal()
        normal =  adjacent_normal_sum/len(self.faces)
        return normal.normalize()

    def min_dirichlet_energy(self):
        num = Vector([0,0,0])
        denom = 0
        for e in self.edges:
            if e.vertices[0]==self:
                neigh = e.vertices[1]
            else:
                neigh = e.vertices[0]
            cot_sum = 0
            for f in e.faces:
                v = f.opposite_vertex(e)
                side1 = neigh.pos - v.pos
                side2 = self.pos  - v.pos
                cot_sum += side1.cot_angle(side2)
            num += cot_sum*neigh.pos
            denom += cot_sum
        res = num/denom

        return res

    def mean_curvature(self):
        v_neighbors = []

        #get adjacent vertices 
        for e in self.edges:
            if e.vertices[0]==self:
                v_neighbors.append(e.vertices[1])
            else:
                v_neighbors.append(e.vertices[0])
        normal =  self.normal()
        #rotate points around origin so normal points to z
        x,y,z = normal.coords
        mag = normal.mag()
        xymag = math.sqrt(x*x + y*y)
        cos_theta = x/xymag
        sin_theta = y/xymag
        cos_phi = z/mag
        sin_phi = xymag/mag
    
        rotation_mat = Matrix([[cos_phi * cos_theta, -cos_phi*sin_theta, -sin_phi],
                        [sin_theta, cos_theta, 0],
                        [sin_theta * sin_phi, -sin_phi * sin_theta, cos_phi]])
    
        crvtr_pts=[]
        for n in v_neighbors:
            max_out_of_plane = 0.1
            translated_neighbor = rotation_mat*(n.pos - self.pos)
            #if our translated neighbor is far enough out of plane, flag self for remeshing
            if translated_neighbor[2] > max_out_of_plane * math.sqrt(translated_neighbor[0]**2 + translated_neighbor[1]**2):
                self.flag = True
            crvtr_pts.append(translated_neighbor)
        
        #least squares fit for a*x**2 + b*x*y + c*y**2 -z
        #initialize coefficients
        Caa = Cbb = Ccc = Cab = Cac = Cbc = Ca = Cb = Cc = 0
        #Sum coefficients of square of difference   (a*x**2 + b*x*y + c*y**2 -z)**2
        for pt in crvtr_pts:
            x,y,z = pt.coords
            Caa += x**4
            Cbb += x**2 * y**2
            Ccc += y**4
            Cab += 2 * x**3 * y
            Cac += 2 * x**2 * y**2
            Cbc += 2 * x * y**3
            Ca  += -2 * x**2 * z
            Cb  += -2 * x * y * z
            Cc  += -2 * y**2 * z
        
        a = -(-(Cac*Cb*Cbc) + Ca*Cbc**2 + 2*Cac*Cbb*Cc - Cab*Cbc*Cc + 2*Cab*Cb*Ccc - 4*Ca*Cbb*Ccc)/(2*(Cac**2*Cbb - Cab*Cac*Cbc + Caa*Cbc**2 + Cab**2*Ccc - 4*Caa*Cbb*Ccc))
        c = -(-(Cab*Cac*Cb) + 2*Ca*Cac*Cbb - Ca*Cab*Cbc + 2*Caa*Cb*Cbc + Cab**2*Cc - 4*Caa*Cbb*Cc)/(2*(Cac**2*Cbb - Cab*Cac*Cbc + Caa*Cbc**2 + Cab**2*Ccc - 4*Caa*Cbb*Ccc))
        return normal, a+c
        
class Edge():
    def __init__(self, v1, v2):
        self.vertices = [v1, v2]
        v1.add_edge(self)
        v2.add_edge(self)
        self.faces = []

    def set_faces(self, faces):
        #edges on trim should have three faces
        #edges not on trim should have two faces
        self.faces = faces
    
    def add_face(self, f):
        self.faces.append(f)

    def remove_face(self, f):
        if f in self.faces:
            self.faces.remove(f)

    def remove_vertex(self, v):
        if v in self.vertices:
            self.vertices.remove(v)

    def clear(self):
        for v in self.vertices:
            v.remove_edge(self)
        for f in self.faces:
            f.remove_edge(self)

    def other(self, face):
        if self.faces[0]==face:
            return self.faces[1]
        elif self.faces[1]==face:
            return self.faces[0]
        else:
            return None

    def length(self):
        return (self.vertices[0].pos - self.vertices[1].pos).mag()

    def dih_cos(self):
        return self.faces[0].normal().cos_angle(self.faces[1].normal())

    def midpoint(self):
        return (self.vertices[0].pos + self.vertices[1].pos)/2

    def sv(self, p):
        if self.vertices[0] == p:
            edge_dir = (self.vertices[1].pos - self.vertices[0].pos).normalize()
        elif self.vertices[1] == p:
            edge_dir = (self.vertices[0].pos - self.vertices[1].pos).normalize()
        else:
            return None
        midpoint = self.midpoint()
        dist_to_mid = (midpoint-p.pos).mag()
        mid_log = math.log(dist_to_mid, 2)
        ce = math.ceil(mid_log)
        fl = math.floor(mid_log)
        ce_point = p.pos + (2**ce)*edge_dir
        fl_point = p.pos + (2**fl)*edge_dir
        if (ce_point - midpoint).mag() > (fl_point - midpoint).mag():
            return Vertex(fl_point)
        else:
            return Vertex(ce_point)

    def is_NLD(self):
        #we will allow a little bit of Non-Locally Delaunay
        angle_sum = 0.0
        for f in self.faces:
            for v in f.vertices:
                if not (v in self.vertices):    
                    angle_sum += (self.vertices[0].pos-v.pos).angle(self.vertices[1].pos-v.pos)
        return (angle_sum/len(self.faces)) >= (math.pi/2)
        
class Face:
    def __init__(self, e1, e2, e3):
        #print e1, e2, e3
        if (e1==e2) or (e2==e3) or (e1==e3):
            print "EQUAL EDGES IN FACE __INIT__"
        self.edges = [e1, e2, e3]
        v1 = list_intersect(e1.vertices, e2.vertices)[0]
        v2 = list_intersect(e2.vertices, e3.vertices)[0]
        v3 = list_intersect(e3.vertices, e1.vertices)[0]
        self.vertices = [v1,v2,v3]
        e1.add_face(self)
        e2.add_face(self)
        e3.add_face(self)
        v1.add_face(self)
        v2.add_face(self)
        v3.add_face(self)

    def clear(self):
        for v in self.vertices:
            v.remove_face(self)
        for e in self.edges:
            e.remove_face(self)

    def normal(self):
        a = self.vertices[0].pos-self.vertices[1].pos
        b = self.vertices[0].pos-self.vertices[2].pos
        return (a^b).normalize()

    def centroid(self):
        return Vertex((self.vertices[0].pos + self.vertices[1].pos + self.vertices[2].pos)/3)

    def remove_edge(self, e):
        if e in self.edges:
            self.edges.remove(e)

    def remove_vertex(self, v):
        if v in self.vertices:
            self.vertices.remove(v)

    def neighbors(self):
        res = []
        for e in self.edges:
            res.append(e.other(self))
        return res
    
    def area(self):
        a = self.edges[0].length()
        b = self.edges[1].length()
        c = self.edges[2].length()        
        s = (a+b+c)/2
        return math.sqrt(s*(s-a)*(s-b)*(s-c))
    
    def next_vertex(self, v):
        if self.vertices[0]==v:
            return self.vertices[1]
        elif self.vertices[1]==v:
            return self.vertices[2]
        elif self.vertices[2]==v:
            return self.vertices[0]
        else:
            return None

    def opposite_vertex(self, e):
        for v in self.vertices:
            if not (v in e.vertices):
                return v
        return None
        

    def opposite_edge(self, v):
        if not (v in self.vertices): return None
        for e in self.edges:
            if not (v in e.vertices):
                return e

class Mesh:
    def __init__(self, vertices, edges, faces):
        self.vertices = vertices
        self.faces = faces
        self.edges = edges
        #self.calc_topology()

    def set_bubble(self, bubble):        
        self.bubble = bubble
                

    def scale(self, factor):
        for v in self.vertices:
            v.pos = v.pos*factor

    def zscale(self, factor):
        for v in self.vertices:
            v.pos.coords[2] = v.pos.coords[2]*factor

    def calc_face_normals(self):
        for face in self.faces:
            face.calc_normal()
    
    def calc_topology():
        #calculate sheets
        faces = self.faces[:]
        sheets = []
        while faces:
            sheet = [faces.pop()]
        #TODO

        #calculate trims
        #calculate singularities    
        #calculate pockets

    def split(self, e, new_vert):

        if len(e.faces)==1:
            self.bubble.wires[0].add_vert(new_vert)
        broken_edges = [Edge(e.vertices[0], new_vert), Edge(e.vertices[1], new_vert)]
        face_edges =[]
        for f in e.faces[:]:
            v = f.opposite_vertex(e)
            face_edge = Edge(v, new_vert)
            face_edges.append(face_edge)
            new_faces = []
            for other_edge in f.edges:
                if other_edge==e: 
                    continue
                elif e.vertices[0] in other_edge.vertices:
                    new_faces.append(Face(broken_edges[0], face_edge, other_edge))
                elif e.vertices[1] in other_edge.vertices:
                    new_faces.append(Face(broken_edges[1], face_edge, other_edge))
                self.edges.remove(other_edge)
                self.edges.append(other_edge)
            f.clear()
            self.faces.remove(f)
            self.faces.extend(new_faces)


        self.edges.extend(face_edges)   
        self.edges.extend(broken_edges)
        e.clear()
        self.edges.remove(e)
        self.vertices.append(new_vert)
        return face_edges
    
    def split_SV(self, e, p):
        if not (p in e.vertices):
            print "SV-splitting vertex must be on edge"
            return None
        return self.split(e, e.sv(p))

    def area(self):
        area = 0
        for f in self.faces:
            area += f.area()
        return area

    def collapse_face(self, f):
        new_vertex = f.centroid()
        neighbors = f.neighbors()
        v_neighbors = []

        next_edge = None
        for v in f.vertices:                
            for v_face in v.faces:
                if v_face == f:
                    continue
                if v_face in neighbors:
                    continue
                v_neighbors.append(v_face)
                
                first_edge = v_face.opposite_edge(v)
                first_face = v_face
                next_face = v_face
                for i in range(3):
                    if v_face.edges[i] == first_edge:
                        next_edge = v_face.edges[(i+1)%3]
        walk_edges = [first_edge]
        walk_faces = [first_face]
        #walk_edges[i] connects walk_verts[i-1] and walk_verts[i]
        walk_vertices = []

        while True:
            pivot = list_intersect(next_edge.vertices, f.vertices)[0]
            next_face = next_edge.other(next_face)
            if next_face==first_face:
                #we are done
                break
            if (next_face in neighbors):
                next_edge = next_face.opposite_edge(pivot)
            else:
                walk_vertices.append(list_intersect(walk_edges[-1].vertices, next_face.vertices)[0])
                walk_faces.append(next_face)
                walk_edges.append(next_face.opposite_edge(pivot))
                next_edge = next_face.opposite_edge(list_intersect(walk_edges[-1].vertices, next_edge.vertices)[0])
        
        walk_vertices.append(list_intersect(walk_edges[-1].vertices, walk_edges[0].vertices)[0])
        num_new = len(walk_edges)

        new_edges = [Edge(walk_v, new_vertex) for walk_v in walk_vertices]
        new_faces = [Face(walk_edges[i], new_edges[i], new_edges[(i-1)%num_new]) for i in range(num_new)]
 
        dying_edges = []        
        for v in f.vertices:
            for e in v.edges:
                if not (e in dying_edges):
                    dying_edges.append(e)

        self.faces.extend(new_faces)
        self.edges.extend(new_edges)
        self.vertices.append(new_vertex)
        #print len(new_faces), len(v_neighbors)        

        f.clear()
        self.faces.remove(f)

        for neigh in neighbors: 
            neigh.clear()
            if neigh in self.faces:
                self.faces.remove(neigh)
        for neigh in v_neighbors : 
            neigh.clear()
            if neigh in self.faces:
                self.faces.remove(neigh)
    
        for v in f.vertices: 
            v.clear()
            self.vertices.remove(v)

        for e in dying_edges: 
            e.clear()
            self.edges.remove(e)              
        
        #print [len(e.faces) for e in new_edges if len(e.faces)!=2]

    def flip_edge(self, e):
        t0 =time.time()
        if len(e.faces) !=2:
            return None
        vs = [None,None,None,None]
        vs[0] = e.vertices[0]
        vs[1] = e.faces[0].opposite_vertex(e)
        vs[2] = e.vertices[1]
        vs[3] = e.faces[1].opposite_vertex(e)
        boundary_edges = [list_intersect(vs[i].edges, vs[(i+1)%4].edges)[0] for i in range(4)]
        #move all the adjacent edges to the end for sorting reasons
        for be in boundary_edges:
            self.edges.remove(be)
            self.edges.append(be)
        new_edge = Edge(vs[1], vs[3])
        
        new_faces = [Face(boundary_edges[2*i], new_edge, boundary_edges[(2*i-1)%4]) for i in range(2)]
       
        boundary_faces = [None,None,None,None]

        if boundary_edges[0].other(e.faces[0]):
            boundary_faces[0] = boundary_edges[0].other(e.faces[0])
            boundary_faces[1] = boundary_edges[1].other(e.faces[0])
            boundary_faces[2] = boundary_edges[2].other(e.faces[1])
            boundary_faces[3] = boundary_edges[3].other(e.faces[1])
        else:
            boundary_faces[0] = boundary_edges[0].other(e.faces[1])
            boundary_faces[1] = boundary_edges[1].other(e.faces[1])
            boundary_faces[2] = boundary_edges[2].other(e.faces[0])
            boundary_faces[3] = boundary_edges[3].other(e.faces[0])    
                    
        self.faces.extend(new_faces)
        for f in e.faces[:]:
            f.clear()
            self.faces.remove(f)
        e.clear()
        self.edges.remove(e)
        self.edges.append(new_edge)
        return new_edge

        
    def first_NLD_edge(self, start):
        for i in range(start, len(self.edges)):
            e = self.edges[i]
            """if len(e.faces)==2:
                if abs(e.faces[0].normal().dot(e.faces[1].normal())) < .8:
                    return self.edges[i], i"""
            if e.is_NLD():
                return self.edges[i], i
        return None, None

    def break_sharp_edges(self):
        count =0
        for e in self.edges[:]:
            if len(e.faces)==2:
                normals=[f.normal() for f in e.faces]
                if abs(normals[0].dot(normals[1])) < .9:
                    count += 1
                    splitter0 = e.faces[0].opposite_edge(e.vertices[0])
                    splitter1 = e.faces[1].opposite_edge(e.vertices[1])
                    self.split(splitter0, Vertex(splitter0.midpoint()))
                    self.split(splitter1, Vertex(splitter1.midpoint()))
        return count


    def remesh(self):
        #get rid of high degree vertices

        for v in self.vertices:
            while len(v.edges)>14:
                e_copy = v.edges[:]
                for i in range(len(e_copy)):
                    if i%2:
                        self.flip_edge(e_copy[i])

        original_vertices = self.vertices[:]
        original_edges = self.edges[:]
        non_physical_edges = []


        
        non_NLD_count = flip_count = split_count = original_non_NLD_count = 0
        e, i = self.first_NLD_edge(0)
        while e:
            #print i, len(self.edges)
            non_NLD_count +=1
            if (e in original_edges): 
                original_non_NLD_count +=1
            else:
                e, i = self.first_NLD_edge(i+1)
                continue
                
            if (e in non_physical_edges) and (len(e.faces)==2):
                non_physical_edges.remove(e)
                non_physical_edges.append(self.flip_edge(e))
                flip_count += 1
            else:
                split_count += 1
                if e.vertices[0] in original_vertices:
                    non_physical_edges.extend(self.split_SV(e, e.vertices[0]))           
                else:
                    non_physical_edges.extend(self.split_SV(e, e.vertices[1]))               

            e, i = self.first_NLD_edge(i-3) 
                    
        print 'flips', flip_count, '  splits', split_count
        print 'there were', non_NLD_count, 'non NLD edges. ', original_non_NLD_count , 'of them were original.'




