'''
Created on Mar 28, 2011

@author: Ari Kardasis
'''

SCREEN_SIZE = (800,600)

from OpenGL.GL import *
from OpenGL.GLU import *
from OpenGL.GLE import *
from pygame.locals import *
from mesh import *
from mesh_topology import *
from stepper import *

import mesh_data
import math
import pygame
import time
import threading

from Tkinter import *
import tkFileDialog
 

def resize(width, height):
    glViewport(0, 0, width, height)
    glMatrixMode(GL_PROJECTION)
    glLoadIdentity()
    gluPerspective(45.0, float(width)/height, 0.1, 1000.)
    glMatrixMode(GL_MODELVIEW)
    glLoadIdentity()
    gluLookAt(15,0,12,4,0,4,0,0,1)
       
def init():
    glEnable(GL_DEPTH_TEST)
    glEnable(GL_NORMALIZE);
    glShadeModel(GL_SMOOTH)
    glClearColor(1, 1, 1, 0.0)
    glEnable(GL_LINE_SMOOTH)
    glEnable(GL_BLEND)
    glHint( GL_LINE_SMOOTH_HINT, GL_NICEST );
    glHint( GL_POLYGON_SMOOTH_HINT, GL_NICEST );
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
    glEnable(GL_COLOR_MATERIAL)
    glEnable(GL_LIGHTING)
    glEnable(GL_LIGHT0)        
    glLight(GL_LIGHT0, GL_POSITION,  (1, 0, 10,5))
    #glEnable(GL_TEXTURE_1D)

def load_from_file():
    master = Tk()
    master.withdraw() #hiding tkinter window
    file_path = tkFileDialog.askopenfilename(title="Open file", filetypes=[("obj file",".obj"),("All files",".*")])
    f = open(file_path)
    return Bubble.load_obj(f.read())

def export(m):
    print 'exporting'
    t = time.time()
    fi = open('./output/' + str(t) +'.obj', 'w')
    for v in m.vertices:
        fi.write('v ')
        for i in range(3):
            fi.write(str(v.pos[i])+' ')
        fi.write("\n")
    for f in m.faces:
        fi.write('f ')
        for v in f.vertices:
            fi.write(str(m.vertices.index(v)+1) + ' ')
        fi.write("\n")
    fi.close()
    print 'DONE'
           

def run(bub):
    pygame.init()
    screen = pygame.display.set_mode(SCREEN_SIZE,HWSURFACE|OPENGL|DOUBLEBUF)
    resize(*SCREEN_SIZE)
    init()
    rendermode = 1
    turntable = 0
    dAngle = 1
    stpr = Stepper(bub)

    keep_running = True
    keep_stepping = False
    do_draw_wires = True
    do_draw_surfs = False
    is_NLD_dark = False

    count = 0

    #navaigation init
    rx, ry = (0,0)
    tx, ty = (0,0)
    zpos = 5
    rotate = move = filming = is_threading = False

    while keep_running: 
        count += 1
        turntable += dAngle
        glRotate(dAngle, 0,0,1)
        for event in pygame.event.get():
            if event.type == QUIT:
                return
            elif event.type == KEYUP and event.key == K_ESCAPE:
                return
            elif event.type == KEYUP and event.key == K_w:
                do_draw_wires = not do_draw_wires
            elif event.type == KEYUP and event.key == K_s:
                do_draw_surfs = not do_draw_surfs
            elif event.type == KEYUP and event.key == K_n:
                is_NLD_dark = not is_NLD_dark
            elif event.type == KEYUP and event.key == K_SPACE:
                keep_stepping = not keep_stepping
            elif event.type == KEYUP and event.key == K_t:
                is_threading = not is_threading
            elif event.type == KEYUP and (event.key == K_KP_ENTER or event.key == K_RETURN):
                if is_threading:
                    thread = threading.Thread(target=stpr.step)
                    thread.start()
                else:
                    stpr.step()
                
            elif event.type == KEYUP and event.key == K_f:
                filming = not filming

            #loading
            elif event.type == KEYUP and event.key == K_l:
                bub = load_from_file()
                stpr = Stepper(bub)
            elif event.type == KEYUP and event.key == K_u:
                bub.mesh.scale(1.5)
                print 'scaled.  area:', bub.mesh.area()
            elif event.type == KEYUP and event.key == K_j:
                bub.mesh.scale(0.7)
                print 'scaled.  area:', bub.mesh.area()
            elif event.type == KEYUP and event.key == K_z:
                bub.mesh.zscale(0.7)
                print 'scaled.  area:', bub.mesh.area()
            elif event.type == KEYUP and event.key == K_a:
                bub.mesh.zscale(1.5)
                print 'scaled.  area:', bub.mesh.area()

            #wirte obj
            elif event.type == KEYUP and event.key == K_e:
                export(bub.mesh)


            #mouse navigation
            elif event.type == MOUSEBUTTONDOWN:
                if event.button == 4: zpos = zpos*.9
                elif event.button == 5: zpos = zpos*1.1
                elif event.button == 1: rotate = True
                elif event.button == 3: move = True
            elif event.type == MOUSEBUTTONUP:
                if event.button == 1: rotate = False
                elif event.button == 3: move = False
            elif event.type == MOUSEMOTION:
                i, j = event.rel
                if rotate:
                    rx += i
                    ry += j
                if move:
                    tx += i
                    ty -= j


        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        glLoadIdentity()
    
        glTranslate(tx/20., ty/20., - zpos)
        glRotate(ry, 1, 0, 0)
        glRotate(rx, 0, 1, 0)

        if do_draw_surfs:
            for face in bub.mesh.faces:
                glColor3f(.7,.5,.9)
                glBegin(GL_TRIANGLE_STRIP)
                for i in range(3):
                    glVertex3f(face.vertices[i].pos[0], face.vertices[i].pos[1], face.vertices[i].pos[2])
                glEnd()

        if do_draw_wires:
            for edge in bub.mesh.edges:
                glColor3f(0,0,0)
                if any([(edge.vertices[0] in w.vertices) for w in bub.wires]) and any([(edge.vertices[1] in w.vertices) for w in bub.wires]):
                    glLineWidth(2)
                else:
                    glLineWidth(1)
                glBegin(GL_LINE_STRIP)
                for i in range(2):
                    glVertex3f(*edge.vertices[i].pos)
                glEnd()
        if keep_stepping:
            stpr.step()
        
        if filming:
            pygame.image.save(screen, "png/"+str(count).zfill(5)+'.png')

        pygame.display.flip()

if __name__ == '__main__':
    bub = Bubble.skirt(50, 2)
    #bub = load_from_file()
    print bub
    run(bub)





